from contextlib import nullcontext
import json
import random
from tqdm import tqdm


DEFAULT_MAX_BYTES = 2e6


class RandomJsonGenerator:
    def __init__(self, max_depth=8, ensure_ascii=False):
        self.max_depth = max_depth
        self.ensure_ascii = ensure_ascii

    def random_list(self, max_bytes=DEFAULT_MAX_BYTES, target_len=None):
        return self._dumps(self._random_list(0, max_bytes, target_len)[0])

    def random_dict(self, max_bytes=DEFAULT_MAX_BYTES, target_len=None):
        return self._dumps(self._random_dict(0, max_bytes)[0])

    def _random_list(
        self, depth=0, max_bytes=DEFAULT_MAX_BYTES, target_len=None
    ):
        l = []
        bytes_here = 2
        if depth >= self.max_depth or bytes_here >= max_bytes:
            return l, bytes_here
        n_elems = target_len or 1 + int(random.random() * 10)
        elem_types = [
            self._random_list,
            self._random_dict,
            self._random_str,
        ] + ([self._random_int, self._random_float] if max_bytes < 10 else [])
        t = tqdm() if depth == 0 else nullcontext()
        t.total = max_bytes
        with t:
            while bytes_here < max_bytes:
                if bytes_here > 2:  # i.e. not first
                    bytes_here += 2  # ", "
                elems_left = n_elems-len(l)
                bytes_left = max_bytes-bytes_here
                elem, elem_bytes = random.choice(elem_types)(
                    depth + 1, bytes_left / (elems_left or -1)
                )
                bytes_here += elem_bytes
                l.append(elem)
                if depth == 0:
                    t.n = bytes_here
                    t.refresh()
            if depth == 0:
                t.n = max_bytes
                t.refresh()
        return l, bytes_here

    def _random_dict(self, depth=0, max_bytes=DEFAULT_MAX_BYTES):
        d = {}
        bytes_here = 2
        if depth >= self.max_depth or bytes_here >= max_bytes:
            return d, bytes_here
        n_elems = 1 + int(random.random() * 10)
        elem_types = [
            self._random_list,
            self._random_dict,
            self._random_str,
        ] + ([self._random_int, self._random_float] if max_bytes < 10 else [])
        t = tqdm() if depth == 0 else nullcontext()
        t.total = max_bytes
        with t:
            while bytes_here < max_bytes:
                key, key_bytes = self._random_str(
                    depth + 1, max(max_bytes / n_elems, 4)
                )
                if key in d:
                    continue
                if bytes_here > 2:  # i.e. not first
                    bytes_here += 2  # ", "
                bytes_here += key_bytes + 2
                value, value_bytes = random.choice(elem_types)(
                    depth + 1, max_bytes / n_elems
                )
                d[key] = value
                bytes_here += value_bytes
                if depth == 0:
                    t.n = bytes_here
                    t.refresh()
            if depth == 0:
                t.n = max_bytes
                t.refresh()
        return d, bytes_here

    def _random_str(self, depth=0, max_bytes=DEFAULT_MAX_BYTES):
        nonsingle = "\b\t\nµæðøæſ"
        choices = (
            " 0123456789"
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz" + nonsingle
        )
        nonsingle = set(nonsingle)
        if max_bytes <= 2:
            return "", 2
        l = []
        bytes_here = 2
        while bytes_here < max_bytes:
            c = random.choice(choices)
            if c not in nonsingle:  # just to speed up
                bytes_here += 1
            else:
                bytes_here += len(self._dumps(c).encode("utf-8")) - 2
            l.append(c)
        return "".join(l), bytes_here

    def _random_int(self, depth=0, max_bytes=DEFAULT_MAX_BYTES):
        n = int(self._random_float(depth, max_bytes, 100000)[0])
        return n, len(str(n))

    def _random_float(self, depth=0, max_bytes=DEFAULT_MAX_BYTES, mult=None):
        if mult is None:
            mult = float(f"1e{int(random.random()*20)}")
        f = random.random() * mult
        return f, len(self._dumps(f))

    def _dumps(self, x):
        return json.dumps(x, ensure_ascii=self.ensure_ascii)



import json
import random
from functools import partial
from pathlib import Path
from tempfile import TemporaryDirectory

import json_stream as js
from contexttimer import Timer
from json_stream.tokenizer import tokenize as pure_python_tokenizer
from json_stream_to_standard_types import to_standard_types
from tqdm import tqdm

import json_stream_rs_tokenizer as jsrs

from .random_json_generator import RandomJsonGenerator


def shuffled(l):
    return random.sample(l, k=len(l))


def main(json_bytes=2e6):
    with TemporaryDirectory() as tmp_dir:
        random_json_file_path = Path(tmp_dir) / "random.json"
        print("generating random json...")
        random_json_str = RandomJsonGenerator().random_list(
            max_bytes=json_bytes, target_len=100
        )
        random_json_size = len(random_json_str.encode("utf-8"))
        print(
            f"generated random json {random_json_file_path} "
            f"with size {random_json_size:.3e} bytes"
        )
        random_json_file_path.write_text(random_json_str, encoding="utf-8")
        results = {"python": {}, "rust": {}, "non-streaming": {}}
        for tokenizer_type, load_fn in shuffled(
            [
                ("python", partial(js.load, tokenizer=pure_python_tokenizer)),
                ("rust", partial(js.load, tokenizer=jsrs.RustTokenizer)),
                ("non-streaming", json.load),
            ]
        ):
            print(f"running with {tokenizer_type} tokenizer")
            with Timer() as t:
                with random_json_file_path.open(encoding="utf-8") as f:
                    l = load_fn(f)
                    parsed = [to_standard_types(x) for x in tqdm(l, total=100)]
            print(f"{tokenizer_type} time: {t.elapsed:.2f} s")
            results[tokenizer_type]["elapsed"] = t.elapsed
            results[tokenizer_type]["parsed"] = parsed
        assert (
            results["python"]["parsed"] == results["rust"]["parsed"]
        ), "BUG: Rust and Py results differ!"
        assert (
            results["non-streaming"]["parsed"] == results["rust"]["parsed"]
        ), "BUG: non-streaming and streaming results differ!"
        speedup = results["python"]["elapsed"] / results["rust"]["elapsed"]
        print(f"speedup: {speedup:.2f}")
    return speedup
